	/*********************************************************************

		m2rtf:	Objective-C to RTF converter

		To use RTF files as input to gcc, go and get Charles Lloyd's
		RTF compiler. If you do not like the style this converter
		produces, well, change this code and do whatever you like
		with it...

		This code is provided ``as is'' and without any waranty.

		Munich, November 1992, Harald Schlangmann

		-----

		m2rtf is a filter, options are:

			-font <fontfamily>
			-size <dimen>
			-tabs <dimen>

		Input will be converted from plain ASCII to RTF, method
		declarations and definitions are typeset bold face like those
		found in NeXT's documentation. Comments, either
		introduced by ``//'' or surrounded by ``/+'' and ``+/''
		(replace ``+'' by ``*'', for sure) will be typeset italics. 
		Furthermore all RTF special characters are escaped. 
		Tabstops are defined equidistant with a gap of <dimen>.
		Default gap is 1440. Default font family is Courier. 
		Default font size is 28.
		I did not read the RTF spec, but the unit seems to be half
		a PostScript point, i.e. 1pt == 0.5RTF unit and
		1inch == 36RTF units.

	 *********************************************************************/

	/*	OK, here are the defaults, you might change:
	 */

	int	size = 28;
	char	*font = "Courier";
	int	tabs = 1440;

	char	*commentprefix = "\\gray333\\i ";
	char 	*commentpostfix = "\n\\gray0\\i0 ";
	
	char	*defprefix = "\\b ";
	char	*defpostfix = "\n\\b0 ";
	
	char	*defvarprefix = "\\i ";
	char	*defvarpostfix = "\\i0 ";
	
	char	*deftypeprefix = "";
	char	*deftypepostfix = "";
	
	char	*keyprefix = "\\b ";
	char	*keypostfix = "\n\\b0 ";
	
	/*	Kind of Startstate
	 */
	 
	int	isName;

%x Definition
%x Comment1
%x Comment2
%x CharSeq1
%x CharSeq2

ID		[a-zA-Z0-9_]+
WS		[ \t\n]
DONTDEF		:{WS}*("("[^)]*")")?{WS}*{ID}

%%
	/*
		Handling of character sequences. Do not treat
		anything inside character sequences special,
		just convert braces and backsplash. Obeye escaped
		quotes (start state CharSeq1).
	 */

<CharSeq2>\'		|
<CharSeq1>\"		{
				BEGIN INITIAL;
				ECHO;
			}

<CharSeq1,CharSeq2>\\\\		{
				fputs("\\\\\\\\",stdout);
			}

<CharSeq2>\\\'		{
				fputs("\\\\'",stdout);
			}

<CharSeq1>\\\"		{
				fputs("\\\\\"",stdout);
			}

<CharSeq1,CharSeq2>[{}\\]	{
				fputs("\\",stdout);
				ECHO;
			}

<CharSeq1,CharSeq2>.	{
				ECHO;
			}

	/*
		Surround definitions conforming to NeXT's
		reference pages.
	 */

<Definition>{ID}	{
				if( isName ) {
				
					/*	Name (or part) of selector
					 */
				
					fputs(defprefix,stdout);
					ECHO;
					fputs(defpostfix,stdout);
					
				} else {
				
					/*	Parameter-Name
					 */
					 
					fputs(defvarprefix,stdout);
					ECHO;
					fputs(defvarpostfix,stdout);
					isName = 1;
				
				}
			}

<Definition>"("[^)]*")"	{		 
				fputs(deftypeprefix,stdout);
				ECHO;
				fputs(deftypepostfix,stdout);
			}
			
<Definition>":"		{
				ECHO;
				isName = 0;
			}

<Definition>"{"		{
				BEGIN INITIAL;
				fputs(defpostfix,stdout);
				fputs("\\{",stdout);
			}

<Definition>;		{
				BEGIN INITIAL;
				fputs(defpostfix,stdout);
				ECHO;
			}

	/*
		Comment typesetting
	 */

<Comment1>\n		{
				BEGIN INITIAL;
				fputs(commentpostfix,stdout);
				fputs("\\\n",stdout);
			}

<Comment2>"*/"		{
				BEGIN INITIAL;
				ECHO;
				fputs(commentpostfix,stdout);
			}

<Comment1,Comment2,Definition>[{}\\\n]		{
				fputs("\\",stdout);
				ECHO;
			}

	/*
		Start the above start states
	 */

^-			{
				BEGIN Definition;
				isName = 1;
				ECHO;
			}

"//"			{
				BEGIN Comment1;
				fputs(commentprefix,stdout);
				ECHO;
			}

"/*"			{
				BEGIN Comment2;
				fputs(commentprefix,stdout);
				ECHO;
			}

\'			{
				BEGIN CharSeq2;
				ECHO;
			}

\"			{
				BEGIN CharSeq1;
				ECHO;
			}

	/*
		Escape some characters conforming to RTF.
	 */

[{}\\\n]		{
				fputs("\\",stdout);
				ECHO;
			}

	/*
		Keyword typesetting
	 */

@implementation|@class	|
@private|@protected	|
@public|@selector	|
@protocol|@defs		|
@encode			|
@interface|@end		|
auto|break|case|char	|
continue|default|do	|
double|else|entry|enum	|
extern|float|for|goto	|
if|int|long|register	|
return|short|sizeof	|
static|struct|switch	|
typedef|union|unsigned	|
void|while		|
volatile|const|inline	|
asm|typeof|signed	{
				fputs(keyprefix,stdout);
				ECHO;
				fputs(keypostfix,stdout);
			}

	/*
		Do not recognize keywords inside identifiers
		and precompiler directives (it's faster
		to do it this way---by the way).
	 */

#?{ID}			|
.			ECHO;

%%

void usage(void) {

	fprintf(stderr,"usage: m2rtf [-font <fontfamily>] [-size <dimen>] "
			"[-tabs <dimen>]\n");
	exit(1);

}

int main(int argc,char **argv) {

	/*	Get arguments
	 */

	int	arg = 1;

	while( arg<argc ) {

		switch( argv[arg][1] ) {

			case 's':	/* size */
				size = atoi(argv[++arg]);
				break;
			case 't':	/* tabs */
				tabs = atoi(argv[++arg]);
				break;
			case 'f':	/* font */
				font = argv[++arg];
				break;
			default:
				usage();
		}

		arg++;
	}

	/*	Print header
	 */

	printf("{\\rtf0\\ansi{\\fonttbl\\f0\\fmodern %s;}\n",font);
	printf("\\paperw14640\n\\paperh13060\n");
	printf("\\margl120\n\\margr120\n");
	printf("\\pard");

	/*	Define tab stops
 	 */

	for( arg=1;arg<16;arg++ )
		printf("\\tx%d",arg*tabs);

	/*	Continue header
	 */

	printf("\\f0\\b0\\i0\\ulnone\\fs%d\\fc0\\cf0\n",size);

	/*	Convert contents
	 */

	yylex();

	/*	Print trailer
	 */

	printf("}\n");
	
	return 0;
}
